/*	$LAAS$ */

/*
 * Copyright (c) 2000,2008 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 *                                                        Fri Aug  4 2006
 */

/* librairies standards */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

/* Les erreurs, structures et numrations locales */
#include "libstereopixel.h"

/********************************************************************
 *
 * moyennes_sigma_ima1 :  partir de l'ima1 passe en paramtre,
 * remplit deux images :
 *
 * - une ima4 float o sont stockes les moyennes des pixels sur la fentre
 * "ni * nj"
 *
 * - une ima4 float o sont stockes les INVERSES des carts types des pixels
 * sur la fentre "ni x nj" (qui s'appelle toujours 'im_sigma' dans la
 * structure, mais ce n'est pas grave)
 *
 * Attention : on balaie une ima1, et on stocke dans deux ima8, qui ne sont
 * pas ordonnes pareil.
 *
 * En gros, voici l'algo : lors d'une premire passe, on remplit deux images
 * intermdiaires : une o sont stockes les sommes des pixels sur une colonne
 * de taille "di" ('im_col'), et une o sont stockes les sommes des carrs
 * des pixels sur une colonne de taille "di" ('im_col_carre').
 *
 * Lors d'une seconde passe, on dtermine l'image des moyennes et des
 * sigmas (en appliquant "sigma = carr de la moyenne moins moyenne des
 * carrs). On dtermine aussi l'image des carts types lors de cette passe.
 *
 * Les paramtres:
 * - images_tmp : contient l'image source ('im_rectifiee') les 2 images
 * intermdiaires ncessaires au calcul, et les images rsultat.
 * - wim_nb_lig, win_nb_col : taille de la fentre de corrlation
 *
 ********************************************************************/

int
moyennes_sigmas (unsigned char*		im_source,
		 SP_IMAGES_TMP*		images_tmp,
		 SP_CORRELATION_PARAMS*	correlation_params)

{
  /* Les diffrentes varaibles que l'on va rcuprer dans la structure
     'correlation_params' */
  int			win_nb_lig;
  int			win_nb_col;

  /* Les diffrentes images que l'on va rcuprer dans la structure
     'images_tmp' */
  unsigned char*	q_im_source;
  short*		q_im_col;
  int*			q_im_col_carre;
  float*		q_im_moyenne;
  float*		q_im_ecart;

  /* Les paramtres inclus dans les structures */
  int             nblig;
  int             nbcol;
  int             ni;
  int             nj;


  unsigned char*  pix_source;
  float*          pix_moyenne;
  float*          pix_ecart;	/* Pixels sur l'image source et les images
				   cibles */

  /* Les images que l'on va crer. Note : en fait, tous les 'nblig' et 'nbcol'
   sont identiques - mais on en definit autant que d'images pour respecter une
   sorte d'homognit */

  /* Les images intermdiaires */
  short*          pix_col;
  int*            pix_col_carre;

  /* Suivent quelques pixels particuliers : 'haut', 'bas', 'gauche' et
     'droite' dsignent leur position par rapport  un pixel courant et la
     fentre considre (colonne, ligne ou fentre carre) */
  unsigned char*  pix_haut_source;
  unsigned char*  pix_bas_source;
  int*            pix_gauche_col_carre;
  int*            pix_droite_col_carre;
  short*          pix_gauche_ecart;
  short*          pix_droite_ecart;


  int       ni2, nj2;		/* Demi-taille de la fentre */
  double    inverse_produit_ni_nj; /* 1.0 / (ni x nj) */

  int       i_max, i_max2, j_max; /* Taille de la fentre sur laquelle
				     on va boucler (on ne s'occuppe pas
				     des bords) */

  int       i, j;		/* Indices pour boucler sur les images */
  int       k, l;		/* Pour boucler sur la fentre (ni x nj) */

  float     somme_nj;		/* Sur une ligne, somme des "nj" pixels */
  float     somme_ni;		/* Sur une colonne, somme des "ni" pixels */
  float     somme_nj_carre;	/* Sur une ligne, somme des carrs des
				   "nj" pixels */
  float     somme_ni_carre;	/* Sur une colonne, somme des carrs
				   des "ni" pixels */

  /* Toutes les variables qui suivent sont des variables d'optimisation */
  int      nj2_P_nblig;
  int      indice_premier_pix_col;

  int      indice_ligne_0_col_nj;
  int      indice_ligne_i_col_nj2;


  /*******************************************************************/

  /* Tests de validit des images dans la structure */
  if (images_tmp->state_intermediaire_moyennes1 != SP_IMAGE_ALLOCATED)
    return stereoError (ERR_SP_IMAGE_TMP_INVALIDE);
  if (images_tmp->state_intermediaire_moyennes2 != SP_IMAGE_ALLOCATED)
    return stereoError (ERR_SP_IMAGE_TMP_INVALIDE);

  if (correlation_params->state != SP_CORRELATION_PARAMS_INITIALIZED)
    return stereoError (ERR_SP_CORRELATION_PARAMS_NOT_INITIALIZED);

  win_nb_lig = correlation_params->win_nb_lig;
  win_nb_col = correlation_params->win_nb_col;

  /* Validit des tailles de la fentre de corrlation */
  if ((win_nb_lig < 1)
      || (win_nb_col < 1)
      || ((win_nb_lig % 2) == 0)
      || ((win_nb_col % 2) == 0))
    return stereoError(ERR_SP_TAILLE_FENETRE_CORRELATION);

  ni = win_nb_lig;
  nj = win_nb_col;

  /* Rcupration des images et tailles dans la structure */
  q_im_source = im_source;
  q_im_col = images_tmp->intermediaire_moyennes1;
  q_im_col_carre = images_tmp->intermediaire_moyennes2;
  q_im_moyenne = images_tmp->im_moyennes;
  q_im_ecart = images_tmp->im_sigmas;

  /* Dcodage des paramtres */
  nblig = images_tmp->images_sizes.nb_lig_final;
  nbcol = images_tmp->images_sizes.nb_col_final;


  /* Initialisation des images intermdiaires */
  memset ( (short*) q_im_col, 0, nblig * nbcol * sizeof (short));
  memset ( (int*) q_im_col_carre, 0, nblig * nbcol * sizeof (int));


  /* Quelques initialisations */
  ni2 = ni / 2;
  nj2 = nj / 2;
  i_max = nblig - ni2 - 1;
  j_max = nbcol - nj2 - 1;
  inverse_produit_ni_nj = 1.0 / ((double) ni * nj);

  /**************************************************************************
   * 1re passe : on remplit 'im_col' et 'im_col_carre'
   **************************************************************************/

  /* On boucle sur toutes les colonnes de l'image source */
  indice_premier_pix_col = ni2 * nbcol;

  for (j = 0; j  < nbcol; j++){
    /* On traite la colonne J */
    /*     indice_premier_pix_col = (j * nblig) + ni2;  : maintenant on
	   incrmente 'indice_premier_pix_col'  la fin de la boucle */

    /* On se met au dbut de l'image source */
    pix_source = q_im_source + j;

    /* On initialise les sommes des "nj" pixels ("somme_nj") et de leurs
       carrs ("somme_nj_carre"). */
    somme_ni_carre = 0;
    somme_ni = 0;
    for (k = 0; k < ni; k++){
      somme_ni += *pix_source;
      somme_ni_carre += ((*pix_source) * (*pix_source));

      pix_source += nbcol;
    }
    /* On se positionne au dbut de la colonne qu'on va traiter dans
       l'image source */
    pix_haut_source = q_im_source + j;
    /* On positionne "pix_bas_source" sur la position actuelle du pointeur
       sur l'image source (qui a t incrment lors de l'initialisation des
       sommes) */
    pix_bas_source = pix_source;

    /* initialize pointers to destination images */
    pix_col_carre = &q_im_col_carre[indice_premier_pix_col];
    pix_col = &q_im_col[indice_premier_pix_col];

    /* On va se dplacer le long de la colonne dans "im_source" (donc on
       descend la colonne dans "im_col_carre" et "im_col"). On dduit les
       valeurs des sommes sur les colonnes (ie. 'somme_ni' et
       'somme_ni_carre', qui vont dans 'im_col' et 'im_col_carre' des valeurs
       juste au dessus selon les schma classique suivant : "en_dessous =
       au_dessus + pix_bas - pix_haut" */
    for (i = ni2; i < i_max; i++) {
      *pix_col_carre = somme_ni_carre;
      *pix_col = somme_ni;

      somme_ni_carre += (((*pix_bas_source) * (*pix_bas_source))
			 - ((*pix_haut_source) * (*pix_haut_source)));
      somme_ni += (*pix_bas_source - *pix_haut_source);

      /* On se dplace d'un pixel vers le bas dans "im_source"... */
      pix_bas_source+=	nbcol;
      pix_haut_source+=	nbcol;
      /* ... et d'un pixel vers le bas dans "im_col_carre" et "im_col" */
      pix_col_carre+=	nbcol;
      pix_col+=		nbcol;
    }
    *pix_col_carre = somme_ni_carre;
    *pix_col = somme_ni;

    indice_premier_pix_col++;
  } /* Les images de sommes sur les colonnes (im_col et im_col_carre) sont
     remplies */

  /**********************************************************************
   *
   * 2nde passe : on dtermine l'image des moyennes et des sigmas (en
   * appliquant "sigma = carr de la moyenne moins moyenne des carrs").
   *
   **********************************************************************/

  /* Initialisation des images de moyenne et carts types des pixels */
  memset ( (float*) q_im_moyenne, 0, nblig * nbcol * sizeof (float));
  memset ( (float*) q_im_ecart, 0, nblig * nbcol * sizeof (float));

  /* loop over columns images */
  i_max2 = nblig - ni2; /* pour ne pas depasser les bords de l'image*/
  indice_ligne_0_col_nj = nj;
  nj2_P_nblig = nj2;

  for ( i = ni2; i < i_max2 ; i++) {
    /* On traite la ligne I */
    pix_col_carre = q_im_col_carre + i * nbcol;
    pix_col = q_im_col + i * nbcol;

    /* On initialise les sommes des "nj x ni" pixels */
    somme_nj_carre = 0;
    somme_nj = 0;
    for (l = 0; l < nj; l++) {

      somme_nj_carre += *pix_col_carre;
      somme_nj += *pix_col;

      pix_col_carre++;
      pix_col++;
    }

    /* On se positionne au dbut de la ligne qu'on va traiter dans toutes
       les images */
    indice_ligne_i_col_nj2 = i*nbcol + nj2;

    pix_gauche_col_carre = q_im_col_carre + i*nbcol;
    pix_droite_col_carre = pix_gauche_col_carre + nj;
    pix_gauche_ecart = q_im_col + i*nbcol;
    pix_droite_ecart = pix_gauche_ecart + nj;

    pix_moyenne = q_im_moyenne + indice_ligne_i_col_nj2;
    pix_ecart = q_im_ecart + indice_ligne_i_col_nj2;

    /* On va dduire les sommes sur les fentres en J (im_moyenne et
       im_ecart)  partir des sommes en J - 1 (toujours le schma classique
       des masques) */

    for (j = nj2; j < j_max; j++) {
      /* Affectation des sommes prcdemment calcules dans les images
	 correspondantes */
      *pix_moyenne = somme_nj * inverse_produit_ni_nj;
      *pix_ecart = (somme_nj_carre * inverse_produit_ni_nj)
		   - ((*pix_moyenne) * (*pix_moyenne))  ;

      /* Calcul des nouvelles sommes */
      somme_nj_carre += (*pix_droite_col_carre - *pix_gauche_col_carre);
      somme_nj += (*pix_droite_ecart - *pix_gauche_ecart);

      /* On se dplace d'un pixel vers la droite */
      pix_moyenne++;
      pix_gauche_col_carre++;
      pix_droite_col_carre++;

      pix_ecart++;
      pix_gauche_ecart++;
      pix_droite_ecart++;
    }

    /* Affectation des sommes prcdemment calcules dans les images
       correspondantes */
    *pix_moyenne = somme_nj * inverse_produit_ni_nj;
    *pix_ecart =  (somme_nj_carre * inverse_produit_ni_nj)
		  - ((*pix_moyenne) * (*pix_moyenne));

  }

  /* Pour conclure : on passe les sigmas en carts types (aprs seuillage
     des plus petites sigmas) */
  pix_ecart = q_im_ecart;
  for (i = 0; i  < nblig; i++){
    for (j = 0; j  < nbcol; j++){
      if (*pix_ecart < 1e-4)
	*pix_ecart = 1e2;
      else
	*pix_ecart = 1.0 / sqrt(*pix_ecart);

      pix_ecart++;
    }
  }
  /* Tout s'est bien pass. On efface les images intermdiaires.... */

  /* Tout baigne */
  images_tmp->state_moyennes = SP_IMAGE_COMPUTED;
  images_tmp->state_sigmas = SP_IMAGE_COMPUTED;

  images_tmp->state_intermediaire_moyennes1 = SP_IMAGE_COMPUTED;
  images_tmp->state_intermediaire_moyennes2 = SP_IMAGE_COMPUTED;

  return (STEREO_OK);
}
